"""Phono3py command option and conf file parser."""

# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import argparse
import os
from typing import Literal

import numpy as np
from phonopy.cui.settings import ConfParser, Settings, fracval


class Phono3pySettings(Settings):
    """Setting parameter container."""

    def __init__(self, load_phono3py_yaml: bool = False):
        """Init method."""
        super().__init__(load_phonopy_yaml=load_phono3py_yaml)
        self.boundary_mfp = 1.0e6  # In micrometer. Value is for avoiding divergence.
        self.conductivity_type = None
        self.constant_averaged_pp_interaction = None
        self.create_forces_fc2 = None
        self.create_forces_fc3 = None
        self.create_forces_fc3_file = None
        self.cutoff_fc3_distance = None
        self.cutoff_pair_distance = None
        self.emulate_v2 = False
        self.grid_addresses = None
        self.grid_points = None
        self.grid_matrix = None
        self.ion_clamped = False
        self.is_bterta = False
        self.is_compact_fc = False
        self.is_fc3_r0_average = False
        self.is_full_pp = False
        self.is_gruneisen = False
        self.is_imag_self_energy = False
        self.is_isotope = False
        self.is_joint_dos = False
        self.is_kappa_star = True
        self.is_lbte = False
        self.is_N_U = False
        self.is_plusminus_displacement_fc2 = "auto"
        self.is_real_self_energy = False
        self.is_reducible_collision_matrix = False
        self.is_spectral_function = False
        self.is_symmetrize_fc2 = False
        self.is_symmetrize_fc3_q = False
        self.is_symmetrize_fc3_r = False
        self.is_tetrahedron_method = False
        self.lapack_zheev_uplo: Literal["L", "U"] = "L"
        self.mass_variances = None
        self.max_freepath = None
        self.num_points_in_batch = None
        self.read_collision = None
        if load_phono3py_yaml:
            self.read_fc2 = True
            self.read_fc3 = True
        else:
            self.read_fc2 = False
            self.read_fc3 = False
        self.read_gamma = False
        self.read_phonon = False
        self.read_pp = False
        self.output_yaml_filename = None
        self.phonon_supercell_matrix = None
        self.pinv_cutoff = 1.0e-8
        self.pinv_solver = 0
        self.pinv_method = 0
        self.pp_conversion_factor = None
        self.random_displacements_fc2 = None
        self.scattering_event_class = None  # scattering event class 1 or 2
        self.sigma_cutoff_width = None
        self.solve_collective_phonon = False
        self.show_symfc_memory_usage = False
        self.symfc_memory_size = None
        self.subtract_forces = None
        self.subtract_forces_fc2 = None
        self.temperatures = None
        self.use_ave_pp = False
        self.use_grg = False
        self.write_collision = False
        self.write_gamma_detail = False
        self.write_gamma = False
        self.write_phonon = False
        self.write_pp = False
        self.write_LBTE_solution = False


class Phono3pyConfParser(ConfParser):
    """Phonopy conf parser.

    Attributes
    ----------
    settings : Phono3pySettings
        Phono3py settings container.
    confs : dict
        Dictionary of settings read from conf file or command options.

    """

    def __init__(
        self,
        filename: str | os.PathLike | None = None,
        args: argparse.Namespace | None = None,
        load_phono3py_yaml: bool = False,
    ):
        """Init method."""
        super().__init__()
        if filename is not None:
            self._read_file(filename)
        if args is not None:
            self._read_options(args)
        self._parse_conf()
        self.settings = Phono3pySettings(load_phono3py_yaml=load_phono3py_yaml)
        self._set_settings(self.settings)

    def _read_options(self, args: argparse.Namespace):
        super()._read_options(args)  # store data in self._confs
        if "phonon_supercell_dimension" in args:
            dim_fc2 = args.phonon_supercell_dimension
            if dim_fc2 is not None:
                self._confs["dim_fc2"] = " ".join(dim_fc2)

        if "boundary_mfp" in args:
            if args.boundary_mfp is not None:
                self._confs["boundary_mfp"] = args.boundary_mfp

        if "const_ave_pp" in args:
            const_ave_pp = args.const_ave_pp
            if const_ave_pp is not None:
                self._confs["const_ave_pp"] = const_ave_pp

        if "create_forces_fc2" in args:
            if args.create_forces_fc2:
                self._confs["create_forces_fc2"] = args.create_forces_fc2

        if "create_forces_fc3" in args:
            if args.create_forces_fc3:
                self._confs["create_forces_fc3"] = args.create_forces_fc3

        if "create_forces_fc3_file" in args:
            if args.create_forces_fc3_file:
                cfc3_file = args.create_forces_fc3_file
                self._confs["create_forces_fc3_file"] = cfc3_file

        if "cutoff_fc3_distance" in args:
            cutoff_fc3 = args.cutoff_fc3_distance
            if cutoff_fc3 is not None:
                self._confs["cutoff_fc3_distance"] = cutoff_fc3

        if "cutoff_pair_distance" in args:
            cutoff_pair = args.cutoff_pair_distance
            if cutoff_pair is not None:
                self._confs["cutoff_pair_distance"] = cutoff_pair

        if "grid_addresses" in args:
            grid_adrs = args.grid_addresses
            if grid_adrs is not None:
                self._confs["grid_addresses"] = " ".join(grid_adrs)

        if "grid_points" in args:
            if args.grid_points is not None:
                self._confs["grid_points"] = " ".join(args.grid_points)

        if "grid_matrix" in args:
            if args.grid_matrix is not None:
                self._confs["grid_matrix"] = " ".join(args.grid_matrix)

        if "ion_clamped" in args:
            if args.ion_clamped:
                self._confs["ion_clamped"] = ".true."
            elif args.ion_clamped is False:
                self._confs["ion_clamped"] = ".false."

        if "is_bterta" in args:
            if args.is_bterta:
                self._confs["bterta"] = ".true."
            elif args.is_bterta is False:
                self._confs["bterta"] = ".false."

        if "is_compact_fc" in args:
            if args.is_compact_fc:
                self._confs["compact_fc"] = ".true."
            elif args.is_compact_fc is False:
                self._confs["compact_fc"] = ".false."

        if "emulate_v2" in args:
            if args.emulate_v2:
                self._confs["emulate_v2"] = ".true."
            elif args.emulate_v2 is False:
                self._confs["emulate_v2"] = ".false."

        if "is_gruneisen" in args:
            if args.is_gruneisen:
                self._confs["gruneisen"] = ".true."
            elif args.is_gruneisen is False:
                self._confs["gruneisen"] = ".false."

        if "is_fc3_r0_average" in args:
            if args.is_fc3_r0_average:
                self._confs["fc3_r0_average"] = ".true."
            elif args.is_fc3_r0_average is False:
                self._confs["fc3_r0_average"] = ".false."

        if "is_full_pp" in args:
            if args.is_full_pp:
                self._confs["full_pp"] = ".true."
            elif args.is_full_pp is False:
                self._confs["full_pp"] = ".false."

        if "is_imag_self_energy" in args:
            if args.is_imag_self_energy:
                self._confs["imag_self_energy"] = ".true."
            elif args.is_imag_self_energy is False:
                self._confs["imag_self_energy"] = ".false."

        if "is_isotope" in args:
            if args.is_isotope:
                self._confs["isotope"] = ".true."
            elif args.is_isotope is False:
                self._confs["isotope"] = ".false."

        if "is_joint_dos" in args:
            if args.is_joint_dos:
                self._confs["joint_dos"] = ".true."
            elif args.is_joint_dos is False:
                self._confs["joint_dos"] = ".false."

        if "no_kappa_star" in args:
            if args.kappa_star:
                self._confs["kappa_star"] = ".true."
            elif args.kappa_star is False:
                self._confs["kappa_star"] = ".false."

        if "is_lbte" in args:
            if args.is_lbte:
                self._confs["lbte"] = ".true."
            elif args.is_lbte is False:
                self._confs["lbte"] = ".false."

        if "is_N_U" in args:
            if args.is_N_U:
                self._confs["N_U"] = ".true."
            elif args.is_N_U is False:
                self._confs["N_U"] = ".false."

        if "is_plusminus_displacements_fc2" in args:
            if args.is_plusminus_displacements_fc2:
                self._confs["pm_fc2"] = ".true."
            elif args.is_plusminus_displacements_fc2 is False:
                self._confs["pm_fc2"] = ".false."

        if "is_real_self_energy" in args:
            if args.is_real_self_energy:
                self._confs["real_self_energy"] = ".true."
            elif args.is_real_self_energy is False:
                self._confs["real_self_energy"] = ".false."

        if "is_reducible_collision_matrix" in args:
            if args.is_reducible_collision_matrix:
                self._confs["reducible_collision_matrix"] = ".true."
            elif args.is_reducible_collision_matrix is False:
                self._confs["reducible_collision_matrix"] = ".false."

        if "is_spectral_function" in args:
            if args.is_spectral_function:
                self._confs["spectral_function"] = ".true."
            elif args.is_spectral_function is False:
                self._confs["spectral_function"] = ".false."

        if "is_symmetrize_fc2" in args:
            if args.is_symmetrize_fc2:
                self._confs["symmetrize_fc2"] = ".true."
            elif args.is_symmetrize_fc2 is False:
                self._confs["symmetrize_fc2"] = ".false."

        if "is_symmetrize_fc3_q" in args:
            if args.is_symmetrize_fc3_q:
                self._confs["symmetrize_fc3_q"] = ".true."
            elif args.is_symmetrize_fc3_q is False:
                self._confs["symmetrize_fc3_q"] = ".false."

        if "is_symmetrize_fc3_r" in args:
            if args.is_symmetrize_fc3_r:
                self._confs["symmetrize_fc3_r"] = ".true."
            elif args.is_symmetrize_fc3_r is False:
                self._confs["symmetrize_fc3_r"] = ".false."

        if "is_tetrahedron_method" in args:
            if args.is_tetrahedron_method:
                self._confs["tetrahedron"] = ".true."
            elif args.is_tetrahedron_method is False:
                self._confs["tetrahedron"] = ".false."

        if "is_wigner_kappa" in args:
            if args.is_wigner_kappa:
                self._confs["conductivity_type"] = "wigner"

        if "is_kubo_kappa" in args:
            if args.is_kubo_kappa:
                self._confs["conductivity_type"] = "kubo"

        if "lapack_zheev_uplo" in args:
            if args.lapack_zheev_uplo is not None:
                self._confs["lapack_zheev_uplo"] = args.lapack_zheev_uplo

        if "mass_variances" in args:
            mass_variances = args.mass_variances
            if mass_variances is not None:
                self._confs["mass_variances"] = " ".join(mass_variances)

        if "max_freepath" in args:
            if args.max_freepath is not None:
                self._confs["max_freepath"] = args.max_freepath

        if "num_points_in_batch" in args:
            num_points_in_batch = args.num_points_in_batch
            if num_points_in_batch is not None:
                self._confs["num_points_in_batch"] = num_points_in_batch

        if "output_yaml_filename" in args:
            if args.output_yaml_filename is not None:
                self._confs["output_yaml_filename"] = args.output_yaml_filename

        if "pinv_cutoff" in args:
            if args.pinv_cutoff is not None:
                self._confs["pinv_cutoff"] = args.pinv_cutoff

        if "pinv_method" in args:
            if args.pinv_method is not None:
                self._confs["pinv_method"] = args.pinv_method

        if "pinv_solver" in args:
            if args.pinv_solver is not None:
                self._confs["pinv_solver"] = args.pinv_solver

        if "pp_conversion_factor" in args:
            pp_conv_factor = args.pp_conversion_factor
            if pp_conv_factor is not None:
                self._confs["pp_conversion_factor"] = pp_conv_factor

        if "random_displacements_fc2" in args:
            rd_fc2 = args.random_displacements_fc2
            if rd_fc2 is not None:
                self._confs["random_displacements_fc2"] = rd_fc2

        if "read_fc2" in args:
            if args.read_fc2:
                self._confs["read_fc2"] = ".true."
            elif args.read_fc2 is False:
                self._confs["read_fc2"] = ".false."

        if "read_fc3" in args:
            if args.read_fc3:
                self._confs["read_fc3"] = ".true."
            elif args.read_fc3 is False:
                self._confs["read_fc3"] = ".false."

        if "read_gamma" in args:
            if args.read_gamma:
                self._confs["read_gamma"] = ".true."
            elif args.read_gamma is False:
                self._confs["read_gamma"] = ".false."

        if "read_phonon" in args:
            if args.read_phonon:
                self._confs["read_phonon"] = ".true."
            elif args.read_phonon is False:
                self._confs["read_phonon"] = ".false."

        if "read_pp" in args:
            if args.read_pp:
                self._confs["read_pp"] = ".true."
            elif args.read_pp is False:
                self._confs["read_pp"] = ".false."

        if "read_collision" in args:
            if args.read_collision is not None:
                self._confs["read_collision"] = args.read_collision

        if "scattering_event_class" in args:
            scatt_class = args.scattering_event_class
            if scatt_class is not None:
                self._confs["scattering_event_class"] = scatt_class

        if "sigma_cutoff_width" in args:
            if args.sigma_cutoff_width is not None:
                self._confs["sigma_cutoff_width"] = args.sigma_cutoff_width

        if "solve_collective_phonon" in args:
            if args.solve_collective_phonon:
                self._confs["collective_phonon"] = ".true."
            elif args.solve_collective_phonon is False:
                self._confs["collective_phonon"] = ".false."

        if "show_symfc_memory_usage" in args:
            if args.show_symfc_memory_usage:
                self._confs["show_symfc_memory_usage"] = ".true."
            elif args.show_symfc_memory_usage is False:
                self._confs["show_symfc_memory_usage"] = ".false."

        if "subtract_forces" in args:
            if args.subtract_forces:
                self._confs["subtract_forces"] = args.subtract_forces

        if "subtract_forces_fc2" in args:
            if args.subtract_forces_fc2:
                self._confs["subtract_forces_fc2"] = args.subtract_forces_fc2

        if "symfc_memory_size" in args:
            if args.symfc_memory_size is not None:
                self._confs["symfc_memory_size"] = args.symfc_memory_size

        if "temperatures" in args:
            if args.temperatures is not None:
                self._confs["temperatures"] = " ".join(args.temperatures)

        if "use_ave_pp" in args:
            if args.use_ave_pp:
                self._confs["use_ave_pp"] = ".true."
            elif args.use_ave_pp is False:
                self._confs["use_ave_pp"] = ".false."

        if "use_grg" in args:
            if args.use_grg:
                self._confs["use_grg"] = ".true."
            elif args.use_grg is False:
                self._confs["use_grg"] = ".false."

        if "write_gamma_detail" in args:
            if args.write_gamma_detail:
                self._confs["write_gamma_detail"] = ".true."
            elif args.write_gamma_detail is False:
                self._confs["write_gamma_detail"] = ".false."

        if "write_gamma" in args:
            if args.write_gamma:
                self._confs["write_gamma"] = ".true."
            elif args.write_gamma is False:
                self._confs["write_gamma"] = ".false."

        if "write_collision" in args:
            if args.write_collision:
                self._confs["write_collision"] = ".true."
            elif args.write_collision is False:
                self._confs["write_collision"] = ".false."

        if "write_phonon" in args:
            if args.write_phonon:
                self._confs["write_phonon"] = ".true."
            elif args.write_phonon is False:
                self._confs["write_phonon"] = ".false."

        if "write_pp" in args:
            if args.write_pp:
                self._confs["write_pp"] = ".true."
            elif args.write_pp is False:
                self._confs["write_pp"] = ".false."

        if "write_LBTE_solution" in args:
            if args.write_LBTE_solution:
                self._confs["write_LBTE_solution"] = ".true."
            elif args.write_LBTE_solution is False:
                self._confs["write_LBTE_solution"] = ".false."

    def _parse_conf(self):
        super()._parse_conf()
        confs = self._confs

        for conf_key in confs.keys():
            # Boolean
            if conf_key in (
                "read_fc2",
                "read_fc3",
                "read_gamma",
                "read_phonon",
                "read_pp",
                "use_ave_pp",
                "use_grg",
                "collective_phonon",
                "write_gamma_detail",
                "write_gamma",
                "write_collision",
                "write_phonon",
                "write_pp",
                "write_LBTE_solution",
                "full_pp",
                "ion_clamped",
                "bterta",
                "compact_fc",
                "emulate_v2",
                "fc3_r0_average",
                "real_self_energy",
                "gruneisen",
                "imag_self_energy",
                "isotope",
                "joint_dos",
                "lbte",
                "N_U",
                "spectral_function",
                "reducible_collision_matrix",
                "show_symfc_memory_usage",
                "symmetrize_fc2",
                "symmetrize_fc3_q",
                "symmetrize_fc3_r",
                "kappa_star",
            ):
                if confs[conf_key].lower() == ".true.":
                    self._set_parameter(conf_key, True)
                elif confs[conf_key].lower() == ".false.":
                    self._set_parameter(conf_key, False)

            # float
            if conf_key in (
                "boundary_mfp",
                "const_ave_pp",
                "cutoff_fc3_distance",
                "cutoff_pair_distance",
                "max_freepath",
                "pinv_cutoff",
                "pp_conversion_factor",
                "sigma_cutoff_width",
                "symfc_memory_size",
            ):
                self._set_parameter(conf_key, float(confs[conf_key]))

            # int
            if conf_key in (
                "pinv_method",
                "pinv_solver",
                "num_points_in_batch",
                "scattering_event_class",
            ):
                self._set_parameter(conf_key, int(confs[conf_key]))

            # string
            if conf_key in (
                "conductivity_type",
                "create_forces_fc3_file",
                "output_yaml_filename",
                "subtract_forces",
                "subtract_forces_fc2",
            ):
                self._set_parameter(conf_key, confs[conf_key])

            # specials
            if conf_key in ("create_forces_fc2", "create_forces_fc3"):
                if isinstance(confs[conf_key], str):
                    fnames = confs[conf_key].split()
                else:
                    fnames = confs[conf_key]
                self._set_parameter(conf_key, fnames)

            if conf_key == "dim_fc2":
                matrix = [int(x) for x in confs["dim_fc2"].split()]
                if len(matrix) == 9:
                    matrix = np.array(matrix).reshape(3, 3)
                elif len(matrix) == 3:
                    matrix = np.diag(matrix)
                else:
                    self.setting_error(
                        "Number of elements of dim-fc2 has to be 3 or 9."
                    )

                if matrix.shape == (3, 3):
                    if np.linalg.det(matrix) < 1:
                        self.setting_error(
                            "Determinant of supercell matrix has " + "to be positive."
                        )
                    else:
                        self._set_parameter("dim_fc2", matrix)

            if conf_key == "grid_addresses":
                vals = [
                    int(x) for x in confs["grid_addresses"].replace(",", " ").split()
                ]
                if len(vals) % 3 == 0 and len(vals) > 0:
                    self._set_parameter("grid_addresses", np.reshape(vals, (-1, 3)))
                else:
                    self.setting_error("Grid addresses are incorrectly set.")

            if conf_key == "grid_points":
                vals = [int(x) for x in confs["grid_points"].replace(",", " ").split()]
                self._set_parameter("grid_points", vals)

            if conf_key == "grid_matrix":
                vals = [int(x) for x in confs["grid_matrix"].replace(",", " ").split()]
                if len(vals) == 9:
                    self._set_parameter("grid_matrix", np.reshape(vals, (3, 3)))
                else:
                    self.setting_error("Grid matrix are incorrectly set.")

            if conf_key == "lapack_zheev_uplo":
                self._set_parameter(
                    "lapack_zheev_uplo", confs["lapack_zheev_uplo"].upper()
                )

            if conf_key == "mass_variances":
                vals = [fracval(x) for x in confs["mass_variances"].split()]
                if len(vals) < 1:
                    self.setting_error("Mass variance parameters are incorrectly set.")
                else:
                    self._set_parameter("mass_variances", vals)

            if conf_key == "read_collision":
                if confs["read_collision"] == "all":
                    self._set_parameter("read_collision", "all")
                else:
                    vals = [int(x) for x in confs["read_collision"].split()]
                    self._set_parameter("read_collision", vals)

            # For multiple T values.
            if conf_key == "temperatures":
                vals = [fracval(x) for x in confs["temperatures"].split()]
                if len(vals) < 1:
                    self.setting_error("Temperatures are incorrectly set.")
                else:
                    self._set_parameter("temperatures", vals)

            if conf_key == "random_displacements_fc2":
                rd = confs["random_displacements_fc2"]
                if rd.lower() == "auto":
                    self._set_parameter("random_displacements_fc2", "auto")
                else:
                    try:
                        self._set_parameter("random_displacements_fc2", int(rd))
                    except ValueError:
                        self.setting_error(f"{conf_key.upper()} is incorrectly set.")

            if conf_key == "pm_fc2":
                if confs["pm_fc2"].lower() == ".false.":
                    self._set_parameter("pm_fc2", False)
                elif confs["pm_fc2"].lower() == ".true.":
                    self._set_parameter("pm_fc2", True)

    def _set_settings(self, settings: Phono3pySettings):
        super()._set_settings(settings)
        params = self._parameters

        # Supercell dimension for fc2
        if "dim_fc2" in params:
            settings.phonon_supercell_matrix = params["dim_fc2"]

        # Boundary mean free path for thermal conductivity calculation
        if "boundary_mfp" in params:
            settings.boundary_mfp = params["boundary_mfp"]

        # Calculate thermal conductivity in BTE-RTA
        if "bterta" in params:
            settings.is_bterta = params["bterta"]

        # Choice of thermal conductivity type
        if "conductivity_type" in params:
            settings.conductivity_type = params["conductivity_type"]

        # Solve collective phonons
        if "collective_phonon" in params:
            settings.solve_collective_phonon = params["collective_phonon"]

        # Compact force constants or full force constants
        if "compact_fc" in params:
            settings.is_compact_fc = params["compact_fc"]

        # Peierls type approximation for squared ph-ph interaction strength
        if "const_ave_pp" in params:
            settings.constant_averaged_pp_interaction = params["const_ave_pp"]

        # Trigger to create FORCES_FC2 and FORCES_FC3
        if "create_forces_fc2" in params:
            settings.create_forces_fc2 = params["create_forces_fc2"]

        if "create_forces_fc3" in params:
            settings.create_forces_fc3 = params["create_forces_fc3"]

        if "create_forces_fc3_file" in params:
            settings.create_forces_fc3_file = params["create_forces_fc3_file"]

        # Cutoff distance of third-order force constants. Elements where any
        # pair of atoms has larger distance than cut-off distance are set zero.
        if "cutoff_fc3_distance" in params:
            settings.cutoff_fc3_distance = params["cutoff_fc3_distance"]

        # Cutoff distance between pairs of displaced atoms used for supercell
        # creation with displacements and making third-order force constants
        if "cutoff_pair_distance" in params:
            settings.cutoff_pair_distance = params["cutoff_pair_distance"]

        # Emulate v2.x behavior
        if "emulate_v2" in params:
            settings.emulate_v2 = params["emulate_v2"]

        # Grid addresses (sets of three integer values)
        if "grid_addresses" in params:
            settings.grid_addresses = params["grid_addresses"]

        # Grid points
        if "grid_points" in params:
            settings.grid_points = params["grid_points"]

        # Grid matrix
        if "grid_matrix" in params:
            settings.mesh_numbers = params["grid_matrix"]

        # Atoms are clamped under applied strain in Gruneisen parameter
        # calculation.
        if "ion_clamped" in params:
            settings.ion_clamped = params["ion_clamped"]

        # Take average in fc3-r2q transformation around three atoms
        if "fc3_r0_average" in params:
            settings.is_fc3_r0_average = params["fc3_r0_average"]

        # Calculate full ph-ph interaction strength for RTA conductivity
        if "full_pp" in params:
            settings.is_full_pp = params["full_pp"]

        # Calculate phonon-Gruneisen parameters
        if "gruneisen" in params:
            settings.is_gruneisen = params["gruneisen"]

        # Calculate imaginary part of self energy
        if "imag_self_energy" in params:
            settings.is_imag_self_energy = params["imag_self_energy"]

        # Calculate lifetime due to isotope scattering
        if "isotope" in params:
            settings.is_isotope = params["isotope"]

        # Calculate joint-DOS
        if "joint_dos" in params:
            settings.is_joint_dos = params["joint_dos"]

        # Sum partial kappa at q-stars
        if "kappa_star" in params:
            settings.is_kappa_star = params["kappa_star"]

        if "lapack_zheev_uplo" in params:
            settings.lapack_zheev_uplo = params["lapack_zheev_uplo"]

        # Calculate thermal conductivity in LBTE with Chaput's method
        if "lbte" in params:
            settings.is_lbte = params["lbte"]

        # Number of frequency points in a batch.
        if "num_points_in_batch" in params:
            settings.num_points_in_batch = params["num_points_in_batch"]

        # Calculate Normal and Umklapp processes
        if "N_U" in params:
            settings.is_N_U = params["N_U"]

        # Plus minus displacement for fc2
        if "pm_fc2" in params:
            settings.is_plusminus_displacement_fc2 = params["pm_fc2"]

        # Solve reducible collision matrix but not reduced matrix
        if "reducible_collision_matrix" in params:
            settings.is_reducible_collision_matrix = params[
                "reducible_collision_matrix"
            ]

        # Symmetrize fc2 by index exchange
        if "symmetrize_fc2" in params:
            settings.is_symmetrize_fc2 = params["symmetrize_fc2"]

        # Symmetrize phonon fc3 by index exchange
        if "symmetrize_fc3_q" in params:
            settings.is_symmetrize_fc3_q = params["symmetrize_fc3_q"]

        # Symmetrize fc3 by index exchange
        if "symmetrize_fc3_r" in params:
            settings.is_symmetrize_fc3_r = params["symmetrize_fc3_r"]

        # Mass variance parameters
        if "mass_variances" in params:
            settings.mass_variances = params["mass_variances"]

        # Maximum mean free path
        if "max_freepath" in params:
            settings.max_freepath = params["max_freepath"]

        # Output yaml filename instead of default filename of phono3py.yaml.
        if "output_yaml_filename" in params:
            settings.output_yaml_filename = params["output_yaml_filename"]

        # Cutoff frequency for pseudo inversion of collision matrix
        if "pinv_cutoff" in params:
            settings.pinv_cutoff = params["pinv_cutoff"]

        # Switch for pseudo-inverse method either taking abs or not.
        if "pinv_method" in params:
            settings.pinv_method = params["pinv_method"]

        # Switch for pseudo-inverse solver
        if "pinv_solver" in params:
            settings.pinv_solver = params["pinv_solver"]

        # Ph-ph interaction unit conversion factor
        if "pp_conversion_factor" in params:
            settings.pp_conversion_factor = params["pp_conversion_factor"]

        # Random displacements for fc2
        if "random_displacements_fc2" in params:
            settings.random_displacements_fc2 = params["random_displacements_fc2"]

        # Calculate real_self_energys
        if "real_self_energy" in params:
            settings.is_real_self_energy = params["real_self_energy"]

            # Read collision matrix and gammas from hdf5
        if "read_collision" in params:
            settings.read_collision = params["read_collision"]

        # Read fc2 from hdf5
        if "read_fc2" in params:
            settings.read_fc2 = params["read_fc2"]

        # Read fc3 from hdf5
        if "read_fc3" in params:
            settings.read_fc3 = params["read_fc3"]

        # Read gammas from hdf5
        if "read_gamma" in params:
            settings.read_gamma = params["read_gamma"]

        # Read phonons from hdf5
        if "read_phonon" in params:
            settings.read_phonon = params["read_phonon"]

        # Read ph-ph interaction strength from hdf5
        if "read_pp" in params:
            settings.read_pp = params["read_pp"]

        # Scattering event class 1 or 2
        if "scattering_event_class" in params:
            settings.scattering_event_class = params["scattering_event_class"]

        # Show symfc memory usage
        if "show_symfc_memory_usage" in params:
            settings.show_symfc_memory_usage = params["show_symfc_memory_usage"]

        # Cutoff width of smearing function (ratio to sigma value)
        if "sigma_cutoff_width" in params:
            settings.sigma_cutoff_width = params["sigma_cutoff_width"]

        # Calculate spectral_functions
        if "spectral_function" in params:
            settings.is_spectral_function = params["spectral_function"]

        # Subtract residual forces to create FORCES_FC2 and FORCES_FC3
        if "subtract_forces" in params:
            settings.subtract_forces = params["subtract_forces"]

        # Subtract residual forces to create FORCES_FC2
        if "subtract_forces_fc2" in params:
            settings.subtract_forces_fc2 = params["subtract_forces_fc2"]

        if "symfc_memory_size" in params:
            settings.symfc_memory_size = params["symfc_memory_size"]

        # Temperatures for scatterings
        if "temperatures" in params:
            settings.temperatures = params["temperatures"]

        # Use averaged ph-ph interaction
        if "use_ave_pp" in params:
            settings.use_ave_pp = params["use_ave_pp"]

        # Use generalized regular grid
        if "use_grg" in params:
            settings.use_grg = params["use_grg"]

        # Write detailed imag-part of self energy to hdf5
        if "write_gamma_detail" in params:
            settings.write_gamma_detail = params["write_gamma_detail"]

        # Write imag-part of self energy to hdf5
        if "write_gamma" in params:
            settings.write_gamma = params["write_gamma"]

        # Write collision matrix and gammas to hdf5
        if "write_collision" in params:
            settings.write_collision = params["write_collision"]

        # Write all phonons on grid points to hdf5
        if "write_phonon" in params:
            settings.write_phonon = params["write_phonon"]

        # Write phonon-phonon interaction amplitudes to hdf5
        if "write_pp" in params:
            settings.write_pp = params["write_pp"]

        # Write direct solution of LBTE to hdf5 files
        if "write_LBTE_solution" in params:
            settings.write_LBTE_solution = params["write_LBTE_solution"]



"""Kubo thermal conductivity RTA class."""

# Copyright (C) 2022 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from phono3py.conductivity.kubo_base import ConductivityKuboMixIn
from phono3py.conductivity.rta_base import ConductivityRTABase
from phono3py.phonon3.interaction import Interaction


class ConductivityKuboRTA(ConductivityKuboMixIn, ConductivityRTABase):
    """Class of Kubo lattice thermal conductivity under RTA."""

    def __init__(
        self,
        interaction: Interaction,
        grid_points=None,
        temperatures=None,
        sigmas=None,
        sigma_cutoff=None,
        is_isotope=False,
        mass_variances=None,
        boundary_mfp=None,  # in micrometer
        use_ave_pp=False,
        is_kappa_star=True,
        gv_delta_q=None,
        is_full_pp=False,
        read_pp=False,
        store_pp=False,
        pp_filename=None,
        is_N_U=False,
        is_gamma_detail=False,
        is_frequency_shift_by_bubble=False,
        log_level=0,
    ):
        """Init method."""
        self._cv_mat = None
        self._gv_mat = None
        self._kappa = None

        super().__init__(
            interaction,
            grid_points=grid_points,
            temperatures=temperatures,
            sigmas=sigmas,
            sigma_cutoff=sigma_cutoff,
            is_isotope=is_isotope,
            mass_variances=mass_variances,
            boundary_mfp=boundary_mfp,
            use_ave_pp=use_ave_pp,
            is_kappa_star=is_kappa_star,
            gv_delta_q=gv_delta_q,
            is_full_pp=is_full_pp,
            read_pp=read_pp,
            store_pp=store_pp,
            pp_filename=pp_filename,
            is_N_U=is_N_U,
            is_gamma_detail=is_gamma_detail,
            is_frequency_shift_by_bubble=is_frequency_shift_by_bubble,
            log_level=log_level,
        )

    def set_kappa_at_sigmas(self):
        """Calculate kappa from ph-ph interaction results."""
        for i_gp, _ in enumerate(self._grid_points):
            frequencies = self._frequencies[self._grid_points[i_gp]]
            for j in range(len(self._sigmas)):
                for k in range(len(self._temperatures)):
                    g_sum = self._get_main_diagonal(i_gp, j, k)
                    for i_band, freq in enumerate(frequencies):
                        if freq < self._pp.cutoff_frequency:
                            self._num_ignored_phonon_modes[j, k] += 1
                            continue
                        self._set_kappa_at_sigmas(
                            j, k, i_gp, i_band, g_sum, frequencies
                        )
        N = self._num_sampling_grid_points
        self._kappa = self._mode_kappa_mat.sum(axis=2).sum(axis=2).sum(axis=2).real / N

    def _set_kappa_at_sigmas(self, j, k, i_gp, i_band, g_sum, frequencies):
        gvm_sum2 = self._gv_mat_sum2[i_gp]
        cvm = self._cv_mat[k, i_gp]
        for j_band, freq in enumerate(frequencies):
            if freq < self._pp.cutoff_frequency:
                return

            g = g_sum[i_band] + g_sum[j_band]
            for i_pair, _ in enumerate(
                ([0, 0], [1, 1], [2, 2], [1, 2], [0, 2], [0, 1])
            ):
                old_settings = np.seterr(all="raise")
                try:
                    self._mode_kappa_mat[j, k, i_gp, i_band, j_band, i_pair] = (
                        cvm[i_band, j_band]
                        * gvm_sum2[i_band, j_band, i_pair]
                        * g
                        / ((frequencies[j_band] - frequencies[i_band]) ** 2 + g**2)
                        * self._conversion_factor
                    )
                except FloatingPointError:
                    # supposed that g is almost 0 and |gv|=0
                    pass
                except Exception:
                    gp = self._grid_points[i_gp]
                    print("=" * 26 + " Warning " + "=" * 26)
                    print(
                        " Unexpected physical condition of ph-ph "
                        "interaction calculation was found."
                    )
                    print(
                        " g=%f at gp=%d, band=%d, freq=%f, band=%d, freq=%f"
                        % (
                            g_sum[i_band],
                            gp,
                            i_band + 1,
                            frequencies[i_band],
                            j_band + 1,
                            frequencies[j_band],
                        )
                    )
                    print("=" * 61)
                np.seterr(**old_settings)

    def _allocate_values(self):
        super()._allocate_values()

        num_band0 = len(self._pp.band_indices)
        num_band = len(self._pp.primitive) * 3
        num_grid_points = len(self._grid_points)
        num_temp = len(self._temperatures)

        self._cv_mat = np.zeros(
            (num_temp, num_grid_points, num_band0, num_band), dtype="double", order="C"
        )
        self._gv_mat = np.zeros(
            (num_grid_points, num_band0, num_band, 3),
            dtype=self._complex_dtype,
            order="C",
        )
        self._gv_mat_sum2 = np.zeros(
            (num_grid_points, num_band0, num_band, 6),
            dtype=self._complex_dtype,
            order="C",
        )

        # kappa and mode_kappa_mat are accessed when all bands exist, i.e.,
        # num_band0==num_band.
        self._kappa = np.zeros(
            (len(self._sigmas), num_temp, 6), dtype="double", order="C"
        )
        self._mode_kappa_mat = np.zeros(
            (len(self._sigmas), num_temp, num_grid_points, num_band, num_band, 6),
            dtype=self._complex_dtype,
            order="C",
        )



"""Tests for direct solution of LBTE."""

import numpy as np
import pytest

from phono3py.api_phono3py import Phono3py


def test_kappa_LBTE(si_pbesol: Phono3py):
    """Test for symmetry reduced collision matrix."""
    if si_pbesol._make_r0_average:
        ref_kappa_P_LBTE = [110.896, 110.896, 110.896, 0, 0, 0]
        ref_kappa_C = [0.166, 0.166, 0.166, 0.000, 0.000, 0.000]
    else:
        ref_kappa_P_LBTE = [111.149, 111.149, 111.149, 0, 0, 0]
        ref_kappa_C = [0.166, 0.166, 0.166, 0.000, 0.000, 0.000]

    si_pbesol.mesh_numbers = [9, 9, 9]
    si_pbesol.init_phph_interaction()
    si_pbesol.run_thermal_conductivity(
        is_LBTE=True,
        temperatures=[
            300,
        ],
        conductivity_type="wigner",
    )
    # kappa = si_pbesol.thermal_conductivity.kappa.ravel()
    kappa_P = si_pbesol.thermal_conductivity.kappa_P_exact.ravel()
    kappa_C = si_pbesol.thermal_conductivity.kappa_C.ravel()
    np.testing.assert_allclose(ref_kappa_P_LBTE, kappa_P, atol=0.5)
    np.testing.assert_allclose(ref_kappa_C, kappa_C, atol=0.02)


@pytest.mark.skip(
    reason=(
        "coherences conductivity is not implemented for "
        "is_reducible_collision_matrix=True"
    )
)
def test_kappa_LBTE_full_colmat(si_pbesol: Phono3py):
    """Test for full collision matrix."""
    si_pbesol_kappa_P_LBTE_redcol = [62.783, 62.783, 62.783, 0, 0, 0]
    si_pbesol_kappa_C_redcol = -1
    si_pbesol.mesh_numbers = [5, 5, 5]
    si_pbesol.init_phph_interaction()
    si_pbesol.run_thermal_conductivity(
        is_LBTE=True,
        temperatures=[
            300,
        ],
        is_reducible_collision_matrix=True,
    )
    kappa_P = si_pbesol.thermal_conductivity.kappa_P_exact.ravel()
    kappa_C = si_pbesol.thermal_conductivity.kappa_C.ravel()
    np.testing.assert_allclose(si_pbesol_kappa_P_LBTE_redcol, kappa_P, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C_redcol, kappa_C, atol=0.02)



"""Lattice thermal conductivity calculation base class with RTA."""

# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import warnings
from abc import abstractmethod

import numpy as np
from phonopy.physical_units import get_physical_units

from phono3py.conductivity.base import ConductivityBase
from phono3py.file_IO import read_pp_from_hdf5
from phono3py.other.tetrahedron_method import get_tetrahedra_relative_grid_address
from phono3py.phonon.grid import (
    get_grid_points_by_rotations,
    get_qpoints_from_bz_grid_points,
)
from phono3py.phonon3.imag_self_energy import ImagSelfEnergy, average_by_degeneracy
from phono3py.phonon3.interaction import Interaction


class ConductivityRTABase(ConductivityBase):
    """Base class of ConductivityRTA*.

    This is a base class of RTA classes.

    """

    def __init__(
        self,
        interaction: Interaction,
        grid_points: np.ndarray | None = None,
        temperatures: list | np.ndarray | None = None,
        sigmas: list | np.ndarray | None = None,
        sigma_cutoff: float | None = None,
        is_isotope: bool = False,
        mass_variances: list | np.ndarray | None = None,
        boundary_mfp: float | None = None,  # in micrometer
        use_ave_pp: bool = False,
        is_kappa_star: bool = True,
        is_full_pp: bool = False,
        read_pp: bool = False,
        store_pp: bool = False,
        pp_filename: float | None = None,
        is_N_U: bool = False,
        is_gamma_detail: bool = False,
        is_frequency_shift_by_bubble: bool = False,
        log_level: int = 0,
    ):
        """Init method."""
        self._is_N_U = is_N_U
        self._is_gamma_detail = is_gamma_detail
        self._is_frequency_shift_by_bubble = is_frequency_shift_by_bubble

        self._gamma_N = None
        self._gamma_U = None
        self._gamma_detail_at_q = None
        self._use_ave_pp = use_ave_pp
        self._use_const_ave_pp = None
        self._num_ignored_phonon_modes = None

        super().__init__(
            interaction,
            grid_points=grid_points,
            temperatures=temperatures,
            sigmas=sigmas,
            sigma_cutoff=sigma_cutoff,
            is_isotope=is_isotope,
            mass_variances=mass_variances,
            boundary_mfp=boundary_mfp,
            is_kappa_star=is_kappa_star,
            is_full_pp=is_full_pp,
            log_level=log_level,
        )

        self._use_const_ave_pp = self._pp.constant_averaged_interaction
        self._read_pp = read_pp
        self._store_pp = store_pp
        self._pp_filename = pp_filename

        if self._temperatures is not None:
            self._allocate_values()

        self._collision = ImagSelfEnergy(
            self._pp, with_detail=(self._is_gamma_detail or self._is_N_U)
        )

    def get_gamma_N_U(self):
        """Return N and U parts of gamma."""
        return (self._gamma_N, self._gamma_U)

    def set_gamma_N_U(self, gamma_N, gamma_U):
        """Set N and U parts of gamma."""
        self._gamma_N = gamma_N
        self._gamma_U = gamma_U

    def get_gamma_detail_at_q(self):
        """Return contribution of each triplet to gamma at current q-point."""
        return self._gamma_detail_at_q

    def get_number_of_ignored_phonon_modes(self):
        """Return number of ignored phonon modes."""
        warnings.warn(
            "Use attribute, number_of_ignored_phonon_modes",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.number_of_ignored_phonon_modes

    @property
    def number_of_ignored_phonon_modes(self):
        """Return number of ignored phonon modes."""
        return self._num_ignored_phonon_modes

    def set_averaged_pp_interaction(self, ave_pp):
        """Set averaged ph-ph interaction."""
        self._averaged_pp_interaction = ave_pp

    @abstractmethod
    def set_kappa_at_sigmas(self):
        """Must be implemented in the inherited class."""
        raise NotImplementedError()

    def _allocate_values(self):
        if self._temperatures is None:
            raise RuntimeError(
                "Temperatures have not been set yet. "
                "Set temperatures before this method."
            )

        num_band0 = len(self._pp.band_indices)
        num_grid_points = len(self._grid_points)
        num_temp = len(self._temperatures)
        if not self._read_gamma:
            self._gamma = np.zeros(
                (len(self._sigmas), num_temp, num_grid_points, num_band0),
                order="C",
                dtype="double",
            )
            if self._is_gamma_detail or self._is_N_U:
                self._gamma_N = np.zeros_like(self._gamma)
                self._gamma_U = np.zeros_like(self._gamma)

        if self._is_isotope:
            self._gamma_iso = np.zeros(
                (len(self._sigmas), num_grid_points, num_band0),
                order="C",
                dtype="double",
            )
        if self._is_full_pp or self._use_ave_pp or self._use_const_ave_pp:
            self._averaged_pp_interaction = np.zeros(
                (num_grid_points, num_band0), order="C", dtype="double"
            )
        self._num_ignored_phonon_modes = np.zeros(
            (len(self._sigmas), num_temp), order="C", dtype="intc"
        )

    def _run_at_grid_point(self):
        i_gp = self._grid_point_count
        self._show_log_header(i_gp)
        grid_point = self._grid_points[i_gp]
        self._set_cv(i_gp, i_gp)
        self._set_velocities(i_gp, i_gp)

        if self._read_gamma:
            if self._use_ave_pp:
                self._collision.set_grid_point(grid_point)
                self._set_gamma_at_sigmas(i_gp)
        else:
            self._collision.set_grid_point(grid_point)
            num_triplets = len(self._pp.get_triplets_at_q()[0])
            if self._log_level:
                print("Number of triplets: %d" % num_triplets, flush=True)

            if (
                self._is_full_pp
                or self._read_pp
                or self._store_pp
                or self._use_ave_pp
                or self._use_const_ave_pp
                or self._is_gamma_detail
            ):
                self._set_gamma_at_sigmas(i_gp)
            else:  # can save memory space
                self._set_gamma_at_sigmas_lowmem(i_gp)

        if self._is_isotope and not self._read_gamma_iso:
            gamma_iso = self._get_gamma_isotope_at_sigmas(i_gp)
            self._gamma_iso[:, i_gp, :] = gamma_iso[:, self._pp.band_indices]

        if self._log_level:
            self._show_log(i_gp)

    def _set_gamma_at_sigmas(self, i):
        for j, sigma in enumerate(self._sigmas):
            self._collision.set_sigma(sigma, sigma_cutoff=self._sigma_cutoff)
            self._collision.run_integration_weights()

            if self._log_level:
                text = "Collisions will be calculated with "
                if sigma is None:
                    text += "tetrahedron method."
                else:
                    text += "sigma=%s" % sigma
                    if self._sigma_cutoff is None:
                        text += "."
                    else:
                        text += "(%4.2f SD)." % self._sigma_cutoff
                print(text)

            if self._read_pp:
                pp, _g_zero = read_pp_from_hdf5(
                    self._pp.mesh_numbers,
                    grid_point=self._grid_points[i],
                    sigma=sigma,
                    sigma_cutoff=self._sigma_cutoff,
                    filename=self._pp_filename,
                    verbose=(self._log_level > 0),
                )
                _, g_zero = self._collision.get_integration_weights()
                if self._log_level:
                    if len(self._sigmas) > 1:
                        print(
                            "Multiple sigmas or mixing smearing and "
                            "tetrahedron method is not supported."
                        )
                if _g_zero is not None and (_g_zero != g_zero).any():
                    raise ValueError("Inconsistency found in g_zero.")
                self._collision.set_interaction_strength(pp)
            elif self._use_ave_pp:
                self._collision.set_averaged_pp_interaction(
                    self._averaged_pp_interaction[i]
                )
            elif self._use_const_ave_pp:
                if self._log_level:
                    print(
                        "Constant ph-ph interaction of %6.3e is used."
                        % self._pp.constant_averaged_interaction
                    )
                self._collision.run_interaction()
                self._averaged_pp_interaction[i] = self._pp.averaged_interaction
            elif j != 0 and (self._is_full_pp or self._sigma_cutoff is None):
                if self._log_level:
                    print("Existing ph-ph interaction is used.")
            else:
                if self._log_level:
                    print("Calculating ph-ph interaction...")
                self._collision.run_interaction(is_full_pp=self._is_full_pp)
                if self._is_full_pp:
                    self._averaged_pp_interaction[i] = self._pp.averaged_interaction

            # Number of triplets depends on q-point.
            # So this is allocated each time.
            if self._is_gamma_detail:
                num_temp = len(self._temperatures)
                self._gamma_detail_at_q = np.empty(
                    ((num_temp,) + self._pp.interaction_strength.shape),
                    dtype="double",
                    order="C",
                )
                self._gamma_detail_at_q[:] = 0

            if self._log_level:
                print("Calculating collisions at temperatures...")
            for k, t in enumerate(self._temperatures):
                self._collision.temperature = t
                self._collision.run()
                self._gamma[j, k, i] = self._collision.imag_self_energy
                if self._is_N_U or self._is_gamma_detail:
                    g_N, g_U = self._collision.get_imag_self_energy_N_and_U()
                    self._gamma_N[j, k, i] = g_N
                    self._gamma_U[j, k, i] = g_U
                if self._is_gamma_detail:
                    self._gamma_detail_at_q[k] = (
                        self._collision.get_detailed_imag_self_energy()
                    )

    def _set_gamma_at_sigmas_lowmem(self, i):
        """Calculate gamma without storing ph-ph interaction strength.

        `svecs` and `multi` below must not be simply replaced by
        `self._pp.primitive.get_smallest_vectors()` because they must be in
        dense format as always so in Interaction class instance.
        `p2s`, `s2p`, and `masses` have to be also given from Interaction
        class instance.

        """
        num_band = len(self._pp.primitive) * 3
        band_indices = self._pp.band_indices
        (
            svecs,
            multi,
            p2s,
            s2p,
            masses,
        ) = self._pp.get_primitive_and_supercell_correspondence()
        triplets_at_q, weights_at_q, _, _ = self._pp.get_triplets_at_q()

        if None in self._sigmas:
            tetrahedra = get_tetrahedra_relative_grid_address(
                self._pp.bz_grid.microzone_lattice
            )

        # It is assumed that self._sigmas = [None].
        temperatures_THz = np.array(
            self._temperatures * get_physical_units().KB / get_physical_units().THzToEv,
            dtype="double",
        )
        for j, sigma in enumerate(self._sigmas):
            self._collision.set_sigma(sigma)
            if self._is_N_U:
                collisions = np.zeros(
                    (2, len(self._temperatures), len(band_indices)),
                    dtype="double",
                    order="C",
                )
            else:
                collisions = np.zeros(
                    (len(self._temperatures), len(band_indices)),
                    dtype="double",
                    order="C",
                )
            import phono3py._phono3py as phono3c

            # True: OpenMP over triplets
            # False: OpenMP over bands
            if self._pp.openmp_per_triplets is None:
                if len(triplets_at_q) > num_band:
                    openmp_per_triplets = True
                else:
                    openmp_per_triplets = False
            else:
                openmp_per_triplets = self._pp.openmp_per_triplets

            if sigma is None:
                phono3c.pp_collision(
                    collisions,
                    np.array(
                        np.dot(tetrahedra, self._pp.bz_grid.P.T),
                        dtype="int64",
                        order="C",
                    ),
                    self._frequencies,
                    self._eigenvectors,
                    triplets_at_q,
                    weights_at_q,
                    self._pp.bz_grid.addresses,
                    self._pp.bz_grid.gp_map,
                    self._pp.bz_grid.store_dense_gp_map * 1 + 1,
                    self._pp.bz_grid.D_diag,
                    self._pp.bz_grid.Q,
                    self._pp.fc3,
                    self._pp.fc3_nonzero_indices,
                    svecs,
                    multi,
                    masses,
                    p2s,
                    s2p,
                    band_indices,
                    temperatures_THz,
                    self._is_N_U * 1,
                    self._pp.symmetrize_fc3q * 1,
                    self._pp.make_r0_average * 1,
                    self._pp.all_shortest,
                    self._pp.cutoff_frequency,
                    openmp_per_triplets * 1,
                )
            else:
                if self._sigma_cutoff is None:
                    sigma_cutoff = -1
                else:
                    sigma_cutoff = float(self._sigma_cutoff)
                phono3c.pp_collision_with_sigma(
                    collisions,
                    sigma,
                    sigma_cutoff,
                    self._frequencies,
                    self._eigenvectors,
                    triplets_at_q,
                    weights_at_q,
                    self._pp.bz_grid.addresses,
                    self._pp.bz_grid.D_diag,
                    self._pp.bz_grid.Q,
                    self._pp.fc3,
                    self._pp.fc3_nonzero_indices,
                    svecs,
                    multi,
                    masses,
                    p2s,
                    s2p,
                    band_indices,
                    temperatures_THz,
                    self._is_N_U * 1,
                    self._pp.symmetrize_fc3q * 1,
                    self._pp.make_r0_average * 1,
                    self._pp.all_shortest,
                    self._pp.cutoff_frequency,
                    openmp_per_triplets * 1,
                )
            col_unit_conv = self._collision.unit_conversion_factor
            pp_unit_conv = self._pp.unit_conversion_factor
            if self._is_N_U:
                col = collisions.sum(axis=0)
                col_N = collisions[0]
                col_U = collisions[1]
            else:
                col = collisions
            for k in range(len(self._temperatures)):
                self._gamma[j, k, i, :] = average_by_degeneracy(
                    col[k] * col_unit_conv * pp_unit_conv,
                    band_indices,
                    self._frequencies[self._grid_points[i]],
                )
                if self._is_N_U:
                    self._gamma_N[j, k, i, :] = average_by_degeneracy(
                        col_N[k] * col_unit_conv * pp_unit_conv,
                        band_indices,
                        self._frequencies[self._grid_points[i]],
                    )
                    self._gamma_U[j, k, i, :] = average_by_degeneracy(
                        col_U[k] * col_unit_conv * pp_unit_conv,
                        band_indices,
                        self._frequencies[self._grid_points[i]],
                    )

    def _show_log(self, i_gp):
        q = get_qpoints_from_bz_grid_points(i_gp, self._pp.bz_grid)
        gp = self._grid_points[i_gp]
        frequencies = self._frequencies[gp][self._pp.band_indices]
        gv = self._conductivity_components.group_velocities[i_gp]

        if self._averaged_pp_interaction is not None:
            ave_pp = self._averaged_pp_interaction[i_gp]
        else:
            ave_pp = None
        self._show_log_value_names()

        if self._log_level > 2:
            self._show_log_values_on_kstar(frequencies, gv, ave_pp, gp, q)
        else:
            self._show_log_values(frequencies, gv, ave_pp)

        print("", end="", flush=True)

    def _show_log_values(self, frequencies, gv, ave_pp):
        if self._is_full_pp or self._use_ave_pp or self._use_const_ave_pp:
            for f, v, pp in zip(frequencies, gv, ave_pp):
                print(
                    "%8.3f   (%8.3f %8.3f %8.3f) %8.3f %11.3e"
                    % (f, v[0], v[1], v[2], np.linalg.norm(v), pp)
                )
        else:
            for f, v in zip(frequencies, gv):
                print(
                    "%8.3f   (%8.3f %8.3f %8.3f) %8.3f"
                    % (f, v[0], v[1], v[2], np.linalg.norm(v))
                )

    def _show_log_values_on_kstar(self, frequencies, gv, ave_pp, gp, q):
        rotation_map = get_grid_points_by_rotations(gp, self._pp.bz_grid)
        for i, j in enumerate(np.unique(rotation_map)):
            for k, (rot, rot_c) in enumerate(
                zip(self._point_operations, self._rotations_cartesian)
            ):
                if rotation_map[k] != j:
                    continue

                print(
                    " k*%-2d (%5.2f %5.2f %5.2f)" % ((i + 1,) + tuple(np.dot(rot, q)))
                )
                if self._is_full_pp or self._use_ave_pp or self._use_const_ave_pp:
                    for f, v, pp in zip(frequencies, np.dot(rot_c, gv.T).T, ave_pp):
                        print(
                            "%8.3f   (%8.3f %8.3f %8.3f) %8.3f %11.3e"
                            % (f, v[0], v[1], v[2], np.linalg.norm(v), pp)
                        )
                else:
                    for f, v in zip(frequencies, np.dot(rot_c, gv.T).T):
                        print(
                            "%8.3f   (%8.3f %8.3f %8.3f) %8.3f"
                            % (f, v[0], v[1], v[2], np.linalg.norm(v))
                        )
        print("")

    def _show_log_value_names(self):
        if self._is_full_pp or self._use_ave_pp or self._use_const_ave_pp:
            text = "Frequency     group velocity (x, y, z)     |gv|       Pqj"
        else:
            text = "Frequency     group velocity (x, y, z)     |gv|"
        if self._conductivity_components.gv_delta_q is None:
            pass
        else:
            text += "  (dq=%3.1e)" % self._conductivity_components.gv_delta_q
        print(text)
